""" 
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 """

import XDHq, XDHqSHRD
from threading import Thread
import threading
import inspect

import signal, sys, os

if sys.version_info[0] == 2:
	import __builtin__ as builtins
else:
	import builtins

# Overriding some functions for the Dev environment.
if XDHqSHRD.isDev():
	if "openpyxl" in sys.modules :
		defaultXLFunction = sys.modules['openpyxl'].load_workbook
		sys.modules['openpyxl'].load_workbook = lambda filename, **kwargs: defaultXLFunction(XDHq.get_asset_filename(filename), **kwargs)
	
	defaultBuiltinsFunction = builtins.open
	builtins.open = lambda filename, *args, **kwargs: defaultBuiltinsFunction(XDHq.get_asset_filename(filename), *args, **kwargs)

def signal_handler(sig, frame):
  sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)


def create_XML(root_tag):
	return XDHq.XML(root_tag)

createXML = create_XML	


def create_HTML(root_tag=""):	# If 'root_tag' is empty, there will be no root tag in the tree.
	return XDHq.XML(root_tag)

createHTML = create_HTML

broadcast_action = XDHq.broadcastAction


def _call(func, userObject, dom, id, action):
	amount = len(inspect.getargspec(func).args)
	args = []

	if ( not(userObject)) :
		amount += 1

	if ( amount == 4 ):
		args.insert(0,action)

	if( amount >= 3 ):
		args.insert(0,id)

	if( amount >= 2 ):
		args.insert(0,dom)

	if( userObject and (amount >= 1 )):
		args.insert(0,userObject)

	return func(*args)


def worker(userCallback,dom,callbacks):
	args=[]
	userObject = None

	if ( userCallback != None ):
		if ( not(inspect.isclass(userCallback)) and len(inspect.getargspec(userCallback).args) == 1 ):
			args.append(dom)

		userObject = userCallback(*args)
	
	while True:
		[action,id] = dom.getAction()

		if dom.isQuitting():
			break

		if action=="" or not "_PreProcess" in callbacks or _call(callbacks["_PreProcess"],userObject, dom, id, action):
			if _call(callbacks[action], userObject, dom, id, action ) and "_PostProcess" in callbacks:
				_call(callbacks["_PostProcess"],userObject, dom, id, action)

	# print("Quitting thread !")

def callback(userCallback,callbacks,instance):
	thread = threading.Thread(target=worker, args=(userCallback, XDHq.DOM(instance), callbacks))
	thread.daemon = True
	thread.start()
	return thread

def launch(callbacks, userCallback = None, headContent = ""):
	XDHq.launch(callback,userCallback,callbacks,headContent)




"""
MIT License

Copyright (c) 2017 Hajime Nakagami<nakagami@gmail.com>
Copyright (c) 2019 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# This is the adaptation of the program found on:
# https://gist.github.com/nakagami/7a7d799bd4bd4ad8fcea96135c4af179

import os, sys, random, itertools, time

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk

EMPTY = 0
BLACK = -1
WHITE = 1

# http://uguisu.skr.jp/othello/5-1.html

WEIGHT_MATRIX = [
  [120, -20, 20, 5, 5, 20, -20, 120],
  [-20, -40, -5, -5, -5, -5, -40, -20],
  [20, -5, 15, 3, 3, 15, -5, 20],
  [5, -5, 3, 3, 3, 3, -5, 5],
  [5, -5, 3, 3, 3, 3, -5, 5],
  [20, -5, 15, 3, 3, 15, -5, 20],
  [-20, -40, -5, -5, -5, -5, -40, -20],
  [120, -20, 20, 5, 5, 20, -20, 120],
]


class Reversi:
  def reset(self):
    self.board = []
    for _ in range(8):
      self.board.append([EMPTY] * 8)

    self.board[3][3] = self.board[4][4] = BLACK
    self.board[4][3] = self.board[3][4] = WHITE

  def __init__(self, orig=None):
    self.reset()

    # copy constructor
    if orig:
      assert isinstance(orig, Reversi)
      for i in range(8):
        for j in range(8):
          self.board[i][j] = orig.board[i][j]

  def count(self, bwe):
    "Count pieces or empty spaces in the board"
    assert bwe in (BLACK, WHITE, EMPTY)
    n = 0
    for i in range(8):
      for j in range(8):
        if self.board[i][j] == bwe:
          n += 1
    return n

  def _has_my_piece(self, bw, x, y, delta_x, delta_y):
    "There is my piece in the direction of (delta_x, delta_y) from (x, y)."
    assert bw in (BLACK, WHITE)
    assert delta_x in (-1, 0, 1)
    assert delta_y in (-1, 0, 1)
    x += delta_x
    y += delta_y

    if x < 0 or x > 7 or y < 0 or y > 7 or self.board[x][y] == EMPTY:
      return False
    if self.board[x][y] == bw:
      return True
    return self._has_my_piece(bw, x, y, delta_x, delta_y)

  def reversible_directions(self, bw, x, y):
    "Can put piece on (x, y) ? Return list of reversible direction tuple"
    assert bw in (BLACK, WHITE)

    directions = []
    if self.board[x][y] != EMPTY:
      return directions

    for d in itertools.product([-1, 1, 0], [-1, 1, 0]):
      if d == (0, 0):
        continue
      nx = x + d[0]
      ny = y + d[1]
      if nx < 0 or nx > 7 or ny < 0 or ny > 7 or self.board[nx][ny] != bw * -1:
        continue
      if self._has_my_piece(bw, nx, ny, d[0], d[1]):
        directions.append(d)
    return directions

  def _reverse_piece(self, bw, x, y, delta_x, delta_y):
    "Reverse pieces in the direction of (delta_x, delta_y) from (x, y) untill bw."
    assert bw in (BLACK, WHITE)

    x += delta_x
    y += delta_y
    assert self.board[x][y] in (BLACK, WHITE)

    if self.board[x][y] == bw:
      return

    self.board[x][y] = bw
    return self._reverse_piece(bw, x, y, delta_x, delta_y)

  def isAllowed(self, x, y, bw):
    return len(self.reversible_directions(bw, x, y)) != 0

  def put(self, x, y, bw):
    """
    True: Put bw's piece on (x, y) and change board status.
    False: Can't put bw's piece on (x, y)
    """

    assert bw in (BLACK, WHITE)
    directions = self.reversible_directions(bw, x, y)
    if len(directions) == 0:
      return False
    self.board[x][y] = bw
    for delta in directions:
      self._reverse_piece(bw, x, y, delta[0], delta[1])
    return True

  def _calc_score(self, bw, weight_matrix):
    assert bw in (BLACK, WHITE)
    my_score = 0
    against_score = 0
    for i in range(8):
      for j in range(8):
        if self.board[i][j] == bw:
          my_score += weight_matrix[i][j]
        elif self.board[i][j] == bw * -1:
          against_score += weight_matrix[i][j]
    return my_score - against_score

  def find_best_position(self, bw, weight_matrix):
    "Return the best next position."
    assert bw in (BLACK, WHITE)

    next_positions = {}
    for i in range(8):
      for j in range(8):
        reversi = Reversi(self)
        if reversi.put(i, j, bw):
          next_positions.setdefault(
            reversi._calc_score(bw, weight_matrix), []
          ).append((i, j))
    if next_positions:
      next_position = random.choice(next_positions[max(next_positions)])
    else:
      next_position = None
    return next_position


# -------------------------------------------------------------------------------

def drawBoard(reversi, dom, prefetch=False):
  board = atlastk.createHTML("tbody")
  for y, row in enumerate(reversi.board):
    board.push_tag("tr")
    for x, r in enumerate(row):
      board.push_tag("td")
      board.put_attribute("id", str(x) + str(y))
      if (r == EMPTY) and (reversi.isAllowed(y, x, reversi.player)):
        board.put_attribute("data-xdh-onevent", "Play")
        if (prefetch == True):
          r = reversi.player
          board.put_attribute(
            "style", "opacity: 0.1; background-color: white;")
      board.put_attribute(
        "class", {EMPTY: 'none', BLACK: 'black', WHITE: 'white'}[r])
      board.pop_tag()
    board.pop_tag()

  dom.inner("board", board)

  dom.set_values({
    "black": reversi.count(BLACK),
    "white": reversi.count(WHITE)
  })


def acConnect(reversi, dom):
  reversi.player = BLACK
  reversi.weight_matrix = WEIGHT_MATRIX
  dom.inner("", open("Main.html").read())
  drawBoard(reversi, dom)
  dom.alert("Welcome to this Reversi (aka Othello) game made with the Atlas toolkit.\n\nYou play against the computer with the black pieces.")


def acPlay(reversi, dom, id):
  xy = [int(id[1]), int(id[0])]

  player = reversi.player
  weight_matrix = reversi.weight_matrix

  if (reversi.put(xy[0], xy[1], player)):
    drawBoard(reversi, dom, False)

    xy = reversi.find_best_position(player * -1, weight_matrix)
    if xy:
      reversi.put(xy[0], xy[1], player * -1)
      time.sleep(1)

    drawBoard(reversi, dom)

  if (reversi.count(EMPTY) == 0 or
    reversi.count(BLACK) == 0 or
      reversi.count(WHITE) == 0):
    if reversi.count(player) > reversi.count(player * -1):
      dom.alert('You win!')
    elif reversi.count(player) < reversi.count(player * -1):
      dom.alert('You lose!')
    else:
      dom.alert('Egality!')


def acNew(reversi, dom):
  reversi.reset()
  drawBoard(reversi, dom)


callbacks = {
  "": acConnect,
  "Play": acPlay,
  "New": acNew
}

atlastk.launch(callbacks, Reversi, open("Head.html").read())



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""

  This file comes from https://www.pygame.org.

  It needs PyGame to be installed, and was modified to be controlled
  by a web interface thanks to the Atlas toolkit (http://atlastk.org).
   
"""


import random, math, pygame, os, sys
from pygame.locals import *
from threading import Thread
import threading

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk

#constants
WINSIZE = [640, 480]
WINCENTER = [320, 240]
NUMSTARS = 150
STEP = 20

def init_star():
  "creates new star values"
  dir = random.randrange(100000)
  velmult = random.random()*.6+.4
  vel = [math.sin(dir) * velmult, math.cos(dir) * velmult]
  return vel, WINCENTER[:]


def initialize_stars():
  "creates a new starfield"
  stars = []
  for x in range(NUMSTARS):
    star = init_star()
    vel, pos = star
    steps = random.randint(0, WINCENTER[0])
    pos[0] = pos[0] + (vel[0] * steps)
    pos[1] = pos[1] + (vel[1] * steps)
    vel[0] = vel[0] * (steps * .09)
    vel[1] = vel[1] * (steps * .09)
    stars.append(star)
  move_stars(stars)
  return stars
  

def draw_stars(surface, stars, color):
  "used to draw (and clear) the stars"
  for vel, pos in stars:
    pos = (int(pos[0]), int(pos[1]))
    surface.set_at(pos, color)


def move_stars(stars):
  "animate the star values"
  for vel, pos in stars:
    pos[0] = pos[0] + vel[0]
    pos[1] = pos[1] + vel[1]
    if not 0 <= pos[0] <= WINSIZE[0] or not 0 <= pos[1] <= WINSIZE[1]:
      vel[:], pos[:] = init_star()
    else:
      vel[0] = vel[0] * 1.05
      vel[1] = vel[1] * 1.05

def acConnect(dom):
  dom.inner("", open("Main.html").read())

pos = WINCENTER

def acCenter(dom):
  global pos;
  pos = [WINSIZE[0]/2, WINSIZE[1]/2]

def acRight(dom):
  global pos;
  pos = [pos[0] + STEP, pos[1]]

def acEnd(dom):
  global pos;
  pos = [WINSIZE[0], pos[1]]

def acDown(dom):
  global pos;
  pos = [pos[0], pos[1] + STEP]

def acBottom(dom):
  global pos;
  pos = [pos[0], WINSIZE[1]]

def acLeft(dom):
  global pos;
  pos = [pos[0] - STEP, pos[1]]

def acBegin(dom):
  global pos;
  pos = [0,pos[1]]

def acUp(dom):
  global pos;
  pos = [pos[0], pos[1] - STEP]

def acTop(dom):
  global pos;
  pos = [pos[0], 0]

def acToggle(dom):
  dom.toggle_class("Simple", "hidden")
  dom.toggle_class("Expert", "hidden")

callbacks = {
  "": acConnect,
  "Center": acCenter,
  "Right": acRight,
  "Down": acDown,
  "Left": acLeft,
  "Up": acUp,
  "End": acEnd,
  "Bottom": acBottom,
  "Begin": acBegin,
  "Top": acTop,
  "Toggle": acToggle,
}

def atlas():
  atlastk.launch(callbacks, None, open("Head.html").read())

def main():
  global pos;
  thread = threading.Thread(target=atlas)
  thread.daemon = True
  thread.start()

  pygame.time.wait(2000);

  "This is the starfield code"
  #create our starfield
  random.seed()
  stars = initialize_stars()
  clock = pygame.time.Clock()
  #initialize and prepare screen
  pygame.init()
  screen = pygame.display.set_mode(WINSIZE)
  pygame.display.set_caption('pygame Stars Example')
  white = 255, 240, 200
  black = 20, 20, 40
  screen.fill(black)

  #main game loop
  done = 0
  while not done:
    draw_stars(screen, stars, black)
    move_stars(stars)
    draw_stars(screen, stars, white)
    pygame.display.update()
    for e in pygame.event.get():
      if e.type == QUIT or (e.type == KEYUP and e.key == K_ESCAPE):
        done = 1
        break
      elif e.type == MOUSEBUTTONDOWN and e.button == 1:
        pos = list(e.pos)


    WINCENTER[:] = pos
    clock.tick(50)

main()



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import GPIOq, os, sys, threading

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk

lock = threading.Lock()

availableUserId = 0
currentUserId = None
settings = {}

mappings = {
  GPIOq.D_ODROID_C2:
  sorted({
#		1: {},
#		2: {},
    3: {},
#		4: {},
    5: {},
#		6: {},
    7: {},
#		8: {},
#		9: {},
#		10: {},
    11: {},
    12: {},
    13: {},
#		14: {},
    15: {},
    16: {},
#		17: {},
    18: {},
    19: {},
#		20: {},
    21: {},
    22: {},
    23: {},
    24: {},
#		25: {},
    26: {},
#		27: {},
#		28: {},
    29: {},
#		30: {},
    31: {},
    32: {},
    33: {},
#		34: {},
    35: {},
    36: {},
#		37: {},
#		38: {},
#		39: {},
#		40: {},
  }),
  GPIOq.D_RASPBERRY_PI:
  sorted({
#		1: {},
#		2: {},
    3: {},
#		4: {},
    5: {},
#		6: {},
    7: {},
    8: {},
#		9: {},
    10: {},
    11: {},
    12: {},
    13: {},
#		14: {},
    15: {},
    16: {},
#		17: {},
    18: {},
    19: {},
#		20: {},
    21: {},
    22: {},
    23: {},
    24: {},
#		25: {},
    26: {},
#		27: {},
#		28: {},
    29: {},
#		30: {},
    31: {},
    32: {},
    33: {},
#		34: {},
    35: {},
    36: {},
    37: {},
    38: {},
#		39: {},
    40: {},
  }),
}

mappings[GPIOq.D_TESTING] = mappings[GPIOq.D_ODROID_C2]

mapping = mappings[GPIOq.device]

class Setting:
  MODE = 0
  VALUE = 1
  SELECTED = 2

class Mode:
  IN = GPIOq.M_IN
  OUT = GPIOq.M_OUT
  PWM = GPIOq.M_PWM
  label = {
    IN: "IN",
    OUT: "OUT",
    PWM: "PWM",
  }

def getNewUserId():
  global availableUserId, lock

  lock.acquire()
  userId = availableUserId
  availableUserId += 1
  lock.release()

  return userId

def setCurrentUserId(id):
  global currentUserId, lock

  lock.acquire()
  wasMe = currentUserId == id or currentUserId == None
  currentUserId = id
  lock.release()

  return wasMe

def set(wId,field,value):
  global settings, lock
  lock.acquire()
  settings[wId][field] = value
  lock.release()

def retrieveMode(wId):
  return Mode.IN

def retrieveValue(wId):
  return 0

def retrieveSelected(wId):
  return False

def retrieveSetting(wId):
  return {
    Setting.MODE: retrieveMode(wId),
    Setting.VALUE: retrieveValue(wId),
    Setting.SELECTED: retrieveSelected(wId),
  }

def retrieveSettings():
  settings = {}

  for key in mapping:
    settings[key] = retrieveSetting(key)
    GPIOq.pinMode(key,Mode.IN)	# Default to IN mode, to avoid short-circuit.

  return settings

def syncSettings():
  global settings, lock

  lock.acquire()
  settings = retrieveSettings()
  lock.release()

def getWId(pattern):
  return int(pattern[pattern.find('.')+1:])

class GPIO:
  def __init__(self):
    self._userId = getNewUserId()
    self._settings = {}
# 	setCurrentUserId(self._userId)	To early ! Must be done at connection !

  def _handleModeButtons(self,dom):
    global mapping
    enable = False
    buttons=[]

    for label in Mode.label:
      buttons.append(Mode.label[label])

    for key in mapping:
      if settings[key][Setting.SELECTED]:
        enable = True
        break

    if enable:
      dom.enable_elements(buttons)
    else:
      dom.disable_elements(buttons)

  def _getSetting(self,wId):
    global settings

    return settings[wId]

  def _getMode(self,wId):
    return self._getSetting(wId)[Setting.MODE]

  def _setMode(self,wId,mode):
    set(wId,Setting.MODE,mode)
    GPIOq.pinMode(wId,1 if mode > 1 else mode)
    if ( mode == Mode.PWM ):
      GPIOq.softPWMCreate(wId)
    set(wId,Setting.VALUE,GPIOq.digitalRead(wId))

  def _getValue(self,wId):
    value = self._getSetting(wId)[Setting.VALUE]

    if ( (value != 0) and (self._getMode(wId) != Mode.PWM) ):
      value = 100

    return value

  def _setValue(self,wId,value):
    set(wId,Setting.VALUE,value)
    mode = self._getMode(wId)
    if ( mode == Mode.IN ):
      sys.exit("Can not set value for a pin in IN mode !")
    elif (mode == Mode.OUT):
      GPIOq.digitalWrite( wId, 1 if value > 0 else 0 )
    elif (mode == Mode.PWM):
      GPIOq.softPWMWrite(wId,value)
    else:
      sys.exit("Unknown mode !")

  def _setSelected(self,wId,value):
    set(wId,Setting.SELECTED, not self._getSetting(wId)[Setting.SELECTED] if value == None else value )

  def _getModeLabel(self,wId):
    return Mode.label[self._getSetting(wId)[Setting.MODE]]

  def _getSelected(self,wId):
    return self._getSetting(wId)[Setting.SELECTED]

  def _buildModeCorpus(self,xml):
    xml.push_tag("Modes")

    for mode in Mode.label:
      xml.push_tag("Mode")
      xml.put_attribute("id", mode)
      xml.put_attribute("Label", Mode.label[mode])
      xml.pop_tag()

    xml.pop_tag()

  def _buildCorpus(self,xml):
    xml.push_tag( "Corpus")

    self._buildModeCorpus(xml)

    xml.pop_tag()

  def _buildXML(self):
    global mapping
    xml = atlastk.createXML("XDHTML")
    self._buildCorpus(xml)
    xml.push_tag("GPIOs")

    for wId in mapping:
      xml.push_tag("GPIO")
      xml.put_attribute( "id", wId)
      xml.put_attribute("Selected", self._getSelected(wId))
      xml.put_attribute("Mode",self._getMode(wId))
      xml.put_attribute("Value",self._getValue(wId))
      xml.pop_tag()

    xml.pop_tag()

    return xml

  def take(self):
    return setCurrentUserId(self._userId)

  def display(self,dom):
    dom.inner("GPIO", self._buildXML(), "GPIO.xsl")
    self._handleModeButtons(dom)

  def setMode(self,dom,wId,mode):
    id = "Value."+str(wId);

    self._setMode(wId, mode)

    dom.set_value("Value." + str(wId),self._getValue(wId))
    dom.setAttribute(id,"value",self._getValue(wId))

    if (mode==Mode.IN):
      dom.disable_element(id)
      dom.setAttribute(id,"step","100")
    elif (mode==Mode.OUT):
      dom.enable_element(id)
      dom.setAttribute(id,"step","100")
    elif (mode==Mode.PWM):
      dom.enable_element(id)
      dom.setAttribute(id,"step","1")
    else:
      sys.exit("???")

  def setValue(self,dom,wId,value):
    self._setValue(wId,value)

  def setSelected(self,dom,wId,value):
    self._setSelected(wId,value)
    self._handleModeButtons(dom)

  def setAllSelected(self,dom,value):
    global mapping

    for key in mapping:
      self._setSelected(int(key),value)

    self.display(dom)	

  def setAllMode(self,dom,mode):
    global mapping,settings

    for key in settings:
      if settings[key][Setting.SELECTED]:
        self._setMode(int(key),mode)

    self.display(dom)
  
def preProcess(GPIO,dom):
  if GPIO.take():
    return True
  else:
    dom.alert("Out of sync! Resynchronizing !")
    GPIO.display(dom)

def acConnect(GPIO,dom):
  dom.inner("", open( "Main.html").read() )
  GPIO.take()
  GPIO.display(dom)

def acSwitchMode(GPIO,dom,id):
  GPIO.setMode(dom,getWId(id),int(dom.get_value(id)))
  
def acChangeValue(GPIO,dom,id):
  GPIO.setValue(dom,getWId(id),int(dom.get_value(id)))

callbacks = {
    "_PreProcess": preProcess,
    "": acConnect,
    "SwitchMode": acSwitchMode,
    "ChangeValue": acChangeValue,
    "Toggle": lambda GPIO, dom, id: GPIO.setSelected(dom,getWId(id),None),
    "All": lambda GPIO, dom: GPIO.setAllSelected(dom, True),
    "None": lambda GPIO, dom: GPIO.setAllSelected(dom, False),
    "Invert": lambda GPIO, dom: GPIO.setAllSelected(dom, None),
    "IN": lambda GPIO, dom: GPIO.setAllMode(dom,Mode.IN),
    "OUT": lambda GPIO, dom: GPIO.setAllMode(dom,Mode.OUT),
    "PWM": lambda GPIO, dom: GPIO.setAllMode(dom,Mode.PWM),
  }

GPIOq.setup()

syncSettings()
    
atlastk.launch(callbacks, GPIO, open("Head.html").read())



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os, sys

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk, html

target=""

"""
From here and up to and including the 'ac_connect' function,
to simplify the writing of the program, there are a lot a quirks
which should not be used by regular developers.
"""

def clean(s,i):
  pattern = f' id="_CGN{i}"'

  while pattern in s:
    s = s.replace(pattern, "")
    i += 1
    pattern = f' id="_CGN{i}"'

  return s.strip(" \n").replace ("    <","<").replace("data-xdh-widget_","data-xdh-widget"),i

def display_code(dom,element,i):
  source = dom.first_child(element);
  code,i = clean(dom.get_value(source),i)
  dom.set_value(dom.next_sibling(source),html.escape(code))

  return i

def ac_connect(dom):
  global target

  dom.inner("", open("Main.html").read())
  current = dom.next_sibling(dom.next_sibling(dom.first_child("")))
  i = 0

  target = ""
  list = "<option disabled selected value> -- Select a widget -- </option>"

  while current != "":
    id = dom.get_attribute(current,"id")
    list += f'<option value="{id}">{id}</option>'
    i = display_code(dom,current,i)
    current = dom.next_sibling(current)

  dom.execute_void("document.querySelectorAll('pre').forEach((block) => {hljs.highlightBlock(block);});")

  dom.set_attribute("ckInput","data-xdh-widget",dom.get_attribute("ckInput","data-xdh-widget_"))
  dom.after("ckInput","")
  dom.inner("List", list)

def ac_select(dom,id):
  global target

  if target:
    dom.add_class(target,"hidden")
  target = dom.get_value(id)
  dom.remove_class(target, "hidden")

def dl_shape(flavors):
  html = atlastk.create_HTML()

  for flavor in flavors:
    html.push_tag("option")
    html.put_attribute("value", flavor)
    html.pop_tag()

  return html

dl_flavors = ["Vanilla", "Chocolate", "Caramel"]  

def ac_dl_submit(dom, id):
  global dl_flavors

  flavor = dom.get_value(id)
  dom.set_value(id, "")
  if not flavor in dl_flavors:
    dl_flavors.append(flavor)
    dl_flavors.sort()
    dom.inner("dlFlavors", dl_shape(dl_flavors))
  dom.set_value("dlOutput", flavor)

def sl_embed(other):
  html = atlastk.create_HTML()

  html.push_tag("option")
  html.put_attribute("selected", "selected")
  html.put_value(other)
#  html.pop_tag()

  return html

def ac_sl_add(dom):
  dom.begin("slOthers", sl_embed(dom.get_value("slInput")))
  dom.set_value("slInput", "")
  dom.focus("slInput")  

callbacks = {
  "": ac_connect,
  "Select": ac_select,

  "btSubmit": lambda dom: dom.alert("Click on button detected!"),

  "pwSubmit": lambda dom, id: dom.set_value("pwOutput", dom.get_value(id)),

  "cbSelect": lambda dom, id: dom.set_value("cbOutput", "{} ({})".format(id, dom.get_value(id))),
  "cbSubmit": lambda dom: dom.alert(str(dom.get_values(["cbBicycle", "cbCar","cbPirogue"]))),

  "rdSelect": lambda dom, id: dom.set_value("rdOutput", id),
  "rdSubmit": lambda dom: dom.alert(str(dom.get_values(["rdEmail", "rdPhone","rdMail"]))),

  "dlSubmit": ac_dl_submit,

  "dtSelect": lambda dom, id: dom.set_value("dtOutput", dom.get_value(id)),

  "clSelect": lambda dom, id: dom.set_value("clOutput", dom.get_value(id)),

  "cpSelect": lambda dom, id: dom.set_value("cpOutput", dom.get_value(id)),

  "rgSlide": lambda dom: dom.set_attribute("rgOutput", "value", (dom.get_value("rgVolume"))),

  "slSelect": lambda dom, id: dom.set_value("slOutput", dom.get_value(id)),
  "slAdd": ac_sl_add,
  "slToggle": lambda dom, id: dom.disable_element("slOthers") if dom.get_value(id) == 'true' else dom.enable_element("slOthers"),

  "ckSubmit": lambda dom, id: dom.set_value("ckOutput", dom.get_value("ckInput")),
}

atlastk.launch(callbacks, None, open("Head.html").read())



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os, sys


D_TESTING = 0
D_ODROID_C2 = 1
D_RASPBERRY_PI = 2

labels = {
  D_TESTING: "Testing",
  D_ODROID_C2: "ODROID-C2",
  D_RASPBERRY_PI: "Raspberry Pi"
}

M_OUT = 0
M_IN = 1
M_PWM = 2

def getModelLabel():
  global labels
  file = "/proc/device-tree/model"
  if os.path.isfile(file):
    return open(file).read()
  else:
    return labels[D_TESTING]

def detectDevice(label):
  global labels
  for key in labels.keys():
    if label.startswith(labels[key]):
      return key

  return D_TESTING

device = detectDevice(getModelLabel())

if ( device == D_TESTING ):
  print( "Unknown device; switching to '" + labels[device] + "' device.")

  def setup():
    pass

  def pinMode(pin,mode):
    pass

  def digitalRead(pin):
    return 0

  def digitalWrite(pin,value):
    pass

  def softPWMCreate(pin):
    pass

  def softPWMWrite(pin,value):
    pass

  def softPWMDestroy(pin):
    pass

elif ( device == D_ODROID_C2 ):
  import wiringpi

  print("'" + labels[device] + "' detected.")

  def setup():
    wiringpi.wiringPiSetupPhys()

  def pinMode(pin,mode):
    mode = (mode + 1) % 2 
    wiringpi.pinMode(pin,mode)

  def digitalRead(pin):
    return wiringpi.digitalRead(pin)

  def digitalWrite(pin,value):
    wiringpi.digitalWrite(pin,value)

  def softPWMCreate(pin):
    wiringpi.softPwmCreate(pin,0,100)

  def softPWMWrite(pin,value):
    wiringpi.softPwmWrite(pin,value)

  def softPWMDestroy(pin):
    pass

elif (device == D_RASPBERRY_PI ):
  import RPi.GPIO as GPIO

  print("'" + labels[device] + "' detected.")

  pwms = {}

  def destroyPWMIfNeeded(pin):
    global pwms
    if (pin in pwms):
      pwms[pin].stop()
      pwms.pop(pin)

  def setup():
    GPIO.setmode(GPIO.BOARD)

  def pinMode(pin,mode):
    destroyPWMIfNeeded(pin)
    GPIO.setup(pin,mode)

  def digitalRead(pin):
    return GPIO.input(pin)

  def digitalWrite(pin,value):
    GPIO.output(pin,value)

  def softPWMCreate(pin):
    global pwms
    destroyPWMIfNeeded(pin)
    GPIO.setup(pin, GPIO.OUT)
    GPIO.output(pin, GPIO.LOW)
    pwms[pin] = GPIO.PWM(pin, 100)
    pwms[pin].start(0)

  def softPWMWrite(pin,value):
    global pwms
    pwms[pin].ChangeDutyCycle(value)

  def softPWMDestroy(pin):
    destroyPWMIfNeeded(pin)
else:
  sys.exit("???")

    







# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os, sys

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk

view_mode_elements = ["Pattern", "CreateButton", "DescriptionToggling", "ViewNotes"]

def put(note, id, xml ):
  xml.push_tag("Note")
  xml.put_attribute("id", id)

  for key in note:
    xml.put_tag_and_value(key, note[key])

  xml.pop_tag()

class Notes:
  def __init__(self):
    self.pattern = ""
    self.hide_descriptions = False
    self.index = 0
    self.notes = [
      {
      'title': '',
      'description': '',
      },
      {
      'title': 'Improve design',
      'description': "Tastes and colors… (aka «CSS aren't my cup of tea…»)",
      },
      {
      'title': 'Fixing bugs',
      'description': "There are bugs ? Really ?",
      },
      {
      'title': 'Implement new functionalities',
      'description': "Although it's almost perfect…, isn't it ?",
      },
    ]

  def handle_descriptions(self,dom):
    if self.hide_descriptions:
      dom.disable_element("ViewDescriptions")
    else:
      dom.enable_element("ViewDescriptions")

  def display_list(self,dom):
    xml = atlastk.create_XML("XDHTML")
    values = {}

    xml.push_tag("Notes")

    for index in range(len(self.notes)):
      if index == 0: # 0 skipped, as it serves as buffer for the new ones.
        continue
      if self.notes[index]['title'][:len(self.pattern)].lower() == self.pattern:
        put(self.notes[index], index, xml)
        values["Description." + str(index)] = self.notes[index]['description']

    dom.inner("Notes", xml, "Notes.xsl")
    dom.set_values(values)
    dom.enable_elements(view_mode_elements)

  def view(self, dom):
    dom.enable_elements(view_mode_elements)
    dom.set_value("Edit." + str(self.index), "")
    self.index = -1

def ac_connect(notes, dom):
  dom.inner("", open( "Main.html").read() )
  notes.display_list(dom)

def ac_toggle_descriptions(notes, dom, id):
  notes.hide_descriptions = dom.get_value(id)=="true"
  notes.handle_descriptions(dom)

def ac_search(notes, dom):
  notes.pattern = dom.get_value("Pattern").lower()
  notes.display_list(dom)

def ac_edit(notes, dom, id):
  index = dom.get_mark(id)
  notes.index = int(index)
  note = notes.notes[notes.index]

  dom.inner("Edit." + index, open( "Note.html").read() )
  dom.set_values({ "Title": note['title'], "Description": note['description'] })
  dom.disable_elements(view_mode_elements)
  dom.focus("Title")

def ac_delete(notes, dom, id):
  if dom.confirm("Are you sure you want to delete this entry?"):
    notes.notes.pop(int(dom.get_mark(id)))
    notes.display_list(dom)

def ac_submit(notes, dom):
  result = dom.get_values(["Title", "Description"])
  title = result["Title"].strip()
  description = result["Description"]

  if title:
    notes.notes[notes.index] = { "title": title, "description": description }

    if notes.index == 0:
      notes.notes.insert(0, { 'title': '', 'description': ''})
      notes.display_list( dom )
    else:
      dom.set_values( { "Title." + str(notes.index): title, "Description." + str(notes.index): description })
      notes.view( dom )
  else:
    dom.alert("Title can not be empty!")
    dom.focus("Title")

def ac_cancel(notes, dom):
  note = notes.notes[notes.index]

  result = dom.get_values(["Title", "Description"])
  title = result["Title"].strip()
  description = result["Description"]

  if (title != note['title']) or (description != note['description']):
    if dom.confirm("Are you sure you want to cancel your modifications?"):
      notes.view( dom )
  else:
    notes.view( dom )

callbacks = {	
  "": ac_connect,
  "ToggleDescriptions": ac_toggle_descriptions,
  "Search": ac_search,
  "Edit": ac_edit,
  "Delete": ac_delete,
  "Submit": ac_submit,
  "Cancel": ac_cancel,
}

atlastk.launch(callbacks, Notes, open("Head.html").read())



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import XDHqSHRD

import inspect, os, socket, sys, threading

if sys.version_info[0] == 2:
	from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer	# For 'repl.it'.
	import XDHqFaaS2
	l = XDHqFaaS2.l
	_writeUInt = XDHqFaaS2.writeUInt
	_writeString = XDHqFaaS2.writeString
	_readUInt = XDHqFaaS2.readUInt
	_getString = XDHqFaaS2.getString
	def _REPLit_convert(str):
		return str
else:
	from http.server import BaseHTTPRequestHandler, HTTPServer	# For 'repl.it'.
	import XDHqFaaS3
	l = XDHqFaaS3.l
	_writeUInt = XDHqFaaS3.writeUInt
	_writeString = XDHqFaaS3.writeString
	_readUInt = XDHqFaaS3.readUInt
	_getString = XDHqFaaS3.getString
	def _REPLit_convert(str):
		return bytes(str,"utf-8")

def _REPLHTML1(url):
	return "<html><body><iframe style=\"border-style: none; width: 100%;height: 100%\" src=\"" + url + "\"</iframe></body></html>"

def _REPLHTML2(url):
	return(
"""
<html>
	<head>
		<script src="https://atlastk.org/xdh/qrcode.min.js"></script>
		<script>
function genQRCode(url) {
"""
+ "\tnew QRCode('qrcode', {width:125, height:125, correctLevel: QRCode.CorrectLevel.L}).makeCode('" + url +"');" +
"""
}
		</script>
	</head>
"""
+ "\t<body onload=\"genQRCode('" + url + "')\">\n"
+
"""
		<div style="display:table; margin: 10px auto 5px auto;">
			<span style="display: table; margin: 15px auto 10px auto;font-style: oblique;">Click or scan this QR code:</span>
			<div style="display: flex; justify-content: space-around;">
"""
+ "\t\t\t\t<a target=\"_blank\" href=\"" + url + "\" alt=\"" + url + "\">" +
"""
 					<div id="qrcode"></div>
				</a>
			</div>
		</div>
	</body>
</html>
"""
)

def _REPLHTML3(url):
	return "<html><body><iframe style=\"border-style: none; width: 100%;height: 100%\" src=\"https://atlastk.org/repl_it.php?url=" + url + "\"</iframe></body></html>"

class _REPLit_class(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(_REPLit_convert(_REPLHTML3(globals()['_REPLit_url'])))

def _REPLit(url):
    global _REPLit_url
    _REPLit_url = url
    server_address = ('', 8080)
    httpd = HTTPServer(server_address, _REPLit_class)
    httpd.handle_request()

_FaaSProtocolLabel = "9efcf0d1-92a4-4e88-86bf-38ce18ca2894"
_FaaSProtocolVersion = "0"
_mainProtocolLabel = "bf077e9f-baca-48a1-bd3f-bf5181a78666"
_mainProtocolVersion = "0"

_writeLock = threading.Lock()
_globalCondition = threading.Condition()

_headContent = ""
_token = ""
_instances = {}

class Instance:
	def __init__(self):
		self.condVar = threading.Condition()
		self.handshakeDone = False
		self.quit = False;
	def set(self,thread,id):
		self.thread = thread
		self.id = id
	def IsHandshakeDone(self):
		if self.handshakeDone:
			return True
		else:
			self.handshakeDone = True
			return False
	def getId(self):
		return self.id
	def wait(self):
		with self.condVar:
			self.condVar.wait()
	def signal(self):
		with self.condVar:
			self.condVar.notify()

def isTokenEmpty():
	return not _token or _token[0] == "&"

def getEnv( name, value= "" ):
	if name in os.environ:
		return os.environ[name].strip()
	else:
		return value.strip()

def writeUInt(value):
	global _socket
	_writeUInt( _socket, value )

def writeSInt(value):
	writeUInt( ( ( -value - 1 ) << 1 ) | 1 if value < 0 else value << 1 )

def writeString(string):
	global _socket
	_writeString(_socket, string)

def writeStrings(strings):
	writeUInt(len(strings))

	for string in strings:
		writeString(string)

def readUInt():
	global _socket
	return _readUInt( _socket)

def readSInt():
	value = readUInt()
	return -( ( value >> 1 ) + 1 ) if ( value & 1 ) else value >> 1

def getString():
	global _socket
	return _getString(_socket)

def getStrings():
	amount = readUInt()
	strings = []

	while amount:
		strings.append(getString())
		amount -= 1

	return strings

def _init():
	global _token, _socket, _wAddr, _wPort, _cgi
	pAddr = "faas1.q37.info"
	pPort = 53700
	_wAddr = ""
	_wPort = ""
	_cgi = "xdh"

	atk = getEnv("ATK").upper()

	if atk == "DEV":
		pAddr = "localhost"
		_wPort = "8080"
		print("\tDEV mode !")
	elif atk == "TEST":
		_cgi = "xdh_"
		print("\tTEST mode!")
	elif atk and ( atk != "REPLIT" ) and (atk != 'NONE'):
		sys.exit("Bad 'ATK' environment variable value : should be 'DEV' or 'TEST' !")

	pAddr = getEnv("ATK_PADDR", pAddr)
	pPort = int(getEnv("ATK_PPORT", str(pPort)))
	_wAddr = getEnv("ATK_WADDR", _wAddr)
	_wPort = getEnv("ATK_WPORT", _wPort)

	if _wAddr == "":
		_wAddr = pAddr

	if _wPort != "":
		_wPort = ":" + _wPort

	if isTokenEmpty():
		_token = getEnv("ATK_TOKEN")

	if _token:
		_token = "&" + _token

	_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	print("Connection to '" + str(pAddr) + ":" + str(pPort) + "'...")
	try:
		_socket.connect((pAddr,pPort))
	except:
		exit("Unable to connect to '" + str(pAddr) + ":" + str(pPort) + "'!")
	else:
		print("Connected to '" + str(pAddr) + ":" + str(pPort) + "'.")
		

def _handshake():
	global _writeLock

	_writeLock.acquire()

	writeString(_FaaSProtocolLabel)
	writeString(_FaaSProtocolVersion)

	_writeLock.release()

	error = getString()

	if error:
		sys.exit(error)

	notification = getString()

	if notification:
		print(notification)

def _ignition():
	global _token
	_writeLock.acquire()

	writeString( _token)
	writeString(_headContent)
	writeString(_wAddr);
	writeString("PYH")

	_writeLock.release()

	_token = getString()

	if isTokenEmpty():
		sys.exit(getString())

	if ( _wPort != ":0" ):
		url = getString()

		print(url)
		print("".rjust(len(url),'^'))
		print("Open above URL in a web browser (click, right click or copy/paste). Enjoy!\n")

		atk = getEnv("ATK").upper()

		if atk == "REPLIT":
			print( "IF THE QR CODE IS NOT DISPLAYED, CLICK THE ABOVE REFRESH BUTTON (see http://q37.info/s/zbgfjtp9).\n")
			_REPLit(url)
		elif atk != 'NONE':
			XDHqSHRD.open(url)

def _serve(callback,userCallback,callbacks ):
	global _writeLock, _globalCondition
	while True:

#		l()

		id = readSInt()
		
		if id == -1:	# Should never happen. 
			sys.exit("Received unexpected undefined command id!")
		if id == -2:    # Value reporting a new session.
			id = readSInt()  # The id of the new session.

			if id in _instances:
				sys.exit("Instance of id '" + id + "' exists but should not !")

			instance = Instance()
			instance.set(callback(userCallback, callbacks, instance),id)
			_instances[id] = instance

			_writeLock.acquire()
			writeSInt(id)
			writeString(_mainProtocolLabel)
			writeString(_mainProtocolVersion)
			_writeLock.release()
		elif id == -3:	# Value instructing that a session is closed.
			id = readSInt();
			if not id in _instances:
				sys.exit("Instance of id '" + str(id) + "' not available for destruction!")
			_instances[id].quit = True
			_instances[id].signal();
			with _globalCondition:
				_globalCondition.wait()
			del _instances[id]	# Seemingly destroy the object and remove the entry too.
		elif not id in _instances:
			sys.exit("Unknown instance of id '" + str(id) + "'!")
		elif not _instances[id].IsHandshakeDone():
			error = getString()

			if error:
				sys.exit(error)

			getString()	# Language. Not handled yet.
		else:
			_instances[id].signal()

			with _globalCondition:
				_globalCondition.wait()

def launch(callback, userCallback,callbacks,headContent):
	global _headContent

	_headContent = headContent

	_init()

	_handshake()

	_ignition()

	_serve(callback,userCallback,callbacks)

def broadcastAction(action,id=""):
	_writeLock.acquire()
	writeSInt(-3)
	writeString(action)
	writeString(id)
	_writeLock.release()	

class DOM_FaaS:
	_firstLaunch = True

	def __init__(self, instance):
		self.instance = instance

	def wait(self):
		self.instance.wait()
		
	def signal(self):
		with _globalCondition:
			_globalCondition.notify()

	def _sendSpecialAction(self, action):
		_writeLock.acquire()
		writeSInt(self.instance.getId())
		writeString(action)
		_writeLock.release()

	def getAction(self):
		if self._firstLaunch:
			self._firstLaunch = False
		else:
			self._sendSpecialAction("#StandBy_1")

		self.wait()

		[id,action]=["",""] if self.instance.quit else [getString(),getString()]

		# The below 'is_quitting()' method MUST be called, or the library will hang. 

		return [action,id]

	def isQuitting(self):
		answer = self.instance.quit

		# Below line were in 'getAction()', but, in case of quitting,
		# 'self.instance' could already be destroyed here.
		self.signal()

		return answer;


	def call(self, command, type, *args):
		_writeLock.acquire()
		writeSInt(self.instance.getId())
		writeString(command)

		writeUInt(type)

		for arg in args:
			if isinstance(arg,str):
				writeUInt(XDHqSHRD.RT_STRING)
				writeString(arg)
			else:
				writeUInt(XDHqSHRD.RT_STRINGS)
				writeStrings(arg)

		writeUInt(XDHqSHRD.RT_VOID)	# To report end of argument list.

		_writeLock.release()

		if type == XDHqSHRD.RT_STRING:
			self.wait()
			string = getString()
			self.signal()
			return string
		elif type == XDHqSHRD.RT_STRINGS:
			self.wait()
			strings = getStrings()
			self.signal()
			return strings
		elif type != XDHqSHRD.RT_VOID:
			sys.exit("Unknown return type !!!")



"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os, sys

os.chdir(os.path.dirname(os.path.realpath(__file__)))
sys.path.append("../../atlastk")

import atlastk

class TodoMVC:
  def __init__(self):
    self.exclude = None
    self.index = -1
    self.todos = []

    if False:	# Set to 'True' for testing purpose.
      self.todos.append({"label": "Todo 1", "completed": False })
      self.todos.append({"label": "Todo 2", "completed": True })

  def items_left(self):
    count = 0

    for index in range(len(self.todos)):
      if not self.todos[index]['completed']:
        count += 1

    return count

  def push(self, todo, id, xml):
    xml.push_tag("Todo")
    xml.put_attribute("id", id)
    xml.put_attribute("completed", "true" if todo['completed'] else "false")
    xml.putValue(todo['label'])
    xml.pop_tag()

  def display_count(self, dom, count):
    text = ""

    if count == 1:
      text = "1 item left"
    elif count != 0:
      text = str(count) + " items left"

    dom.set_value("Count", text)

  def handle_count(self, dom):
    count = self.items_left()

    if count != len(self.todos):
      dom.disable_element("HideClearCompleted")
    else:
      dom.enable_element("HideClearCompleted")

    self.display_count(dom, count)

  def display_todos(self, dom):
    xml = atlastk.create_XML("XDHTML")

    xml.push_tag("Todos")

    for index in range(len(self.todos)):
      todo = self.todos[index]

      if (self.exclude == None) or (todo['completed'] != self.exclude):
        self.push(todo, index, xml)

    xml.pop_tag()

    dom.inner("Todos", xml, "Todos.xsl")
    self.handle_count(dom)

  def submit_new(self, dom):
    value = dom.get_value("Input").strip()
    dom.set_value("Input", "")

    if value:
      self.todos.insert(0, {'label': value, 'completed': False})
      self.display_todos(dom)

  def submit_modification(self, dom):
    index = self.index
    self.index = -1

    value = dom.get_value("Input." + str(index)).strip()
    dom.set_value("Input." + str(index), "")

    if value:
      self.todos[index]['label'] = value

      dom.set_value("Label." + str(index), value)

      dom.remove_classes({"View." + str(index): "hide", "Todo." + str(index): "editing"})
    else:
      self.todos.pop(index)
      self.displayTodos(dom)

def ac_connect(self, dom):
  dom.inner("", open("Main.html").read())
  dom.focus("Input")
  self.display_todos(dom)
  dom.disable_elements(["HideActive", "HideCompleted"])

def ac_destroy(self, dom, id):
  self.todos.pop(int(dom.get_mark(id)))
  self.display_todos(dom)

def ac_toggle(self, dom, id):
  index = int(id)
  self.todos[index]['completed'] = not self.todos[index]['completed']

  dom.toggle_class("Todo." + id, "completed")
  dom.toggle_class("Todo." + id, "active")

  self.handle_count(dom)

def ac_all(self, dom):
  self.exclude = None

  dom.add_class("All", "selected")
  dom.remove_classes({"Active": "selected", "Completed": "selected"})
  dom.disable_elements(["HideActive", "HideCompleted"])

def ac_active(self, dom):
  self.exclude = True

  dom.add_class("Active", "selected")
  dom.remove_classes({"All": "selected", "Completed": "selected"})
  dom.disable_element("HideActive")
  dom.enable_element("HideCompleted")

def ac_completed(self, dom):
  self.exclude = False

  dom.add_class("Completed", "selected")
  dom.remove_classes({"All": "selected", "Active": "selected"})
  dom.disable_element("HideCompleted")
  dom.enable_element("HideActive")

def ac_clear(self, dom):
  index = len(self.todos)

  while index:
    index -= 1

    if self.todos[index]['completed']:
      self.todos.pop(index)

  self.display_todos(dom)

def ac_edit(self, dom, id):
  value = dom.get_mark(id)
  self.index = int(value)

  dom.add_classes({"View." + value: "hide", id: "editing"})
  dom.set_value("Input." + value, self.todos[self.index]['label'])
  dom.focus("Input." + value)

def ac_cancel(self, dom):
  index = str(self.index)
  self.index = -1

  dom.set_value("Input." + index, "")
  dom.remove_classes({"View." + index: "hide", "Todo." + index: "editing"})

callbacks = {
  "": ac_connect,
  "Submit": lambda self, dom: self.submit_new(dom) if self.index == -1 else self.submit_modification(dom),
  "Destroy": ac_destroy,
  "Toggle": ac_toggle,
  "All": ac_all,
  "Active": ac_active,
  "Completed": ac_completed,
  "Clear": ac_clear,
  "Edit": ac_edit,
  "Cancel": ac_cancel,
}

atlastk.launch(callbacks, TodoMVC, open("HeadFaaS.html").read())
